"""@file parent_class.py
CLMModeling abstract class
"""

# pylint: disable=too-many-lines
import warnings

import numpy as np

# functions for the 2h term
from scipy.integrate import quad, simpson
from scipy.interpolate import splev, splrep
from scipy.special import gamma, gammainc, jv

from ..utils import (
    _integ_pzfuncs,
    compute_beta_s_func,
    compute_for_good_redshifts,
    validate_argument,
)
from . import miscentering
from .generic import (
    compute_magnification_bias_from_magnification,
    compute_profile_mass_in_radius,
    compute_rdelta,
    compute_reduced_shear_from_convergence,
    convert_profile_mass_concentration,
)

warnings.filterwarnings("always", module="(clmm).*")


class CLMModeling:
    r"""Object with functions for halo mass modeling

    Attributes
    ----------
    backend: str
        Name of the backend being used
    mdelta: float
        Mass of the profile, in units of :math:`M_\odot`
    cdelta: float
        Concentration of the profile
    massdef : str
        Profile mass definition ("mean", "critical", "virial" - letter case independent)
    delta_mdef : int
        Mass overdensity definition.
    halo_profile_model : str
        Profile model parameterization ("nfw", "einasto", "hernquist" - letter case independent)
    cosmo: Cosmology
        Cosmology object
    hdpm: Object
        Backend object with halo profiles
    mdef_dict: dict
        Dictionary with the definitions for mass
    hdpm_dict: dict
        Dictionary with the definitions for profile
    validate_input: bool
        Validade each input argument
    cosmo_class: type
        Type of used cosmology objects
    z_inf : float
        The value used as infinite redshift
    """

    # pylint: disable=too-many-instance-attributes
    # The disable below is added to avoid a pylint error where it thinks CLMMCosmlogy
    # has duplicates since both have many NotImplementedError functions
    # description of bug at https://github.com/pylint-dev/pylint/issues/7213
    # pylint: disable=duplicate-code

    def __init__(self, validate_input=True, z_inf=1000):
        self.backend = None

        self.__massdef = ""
        self.__delta_mdef = 0
        self.__halo_profile_model = ""

        self.cosmo = None

        self.hdpm = None
        self.mdef_dict = {}
        self.hdpm_dict = {}

        self.validate_input = validate_input
        self.cosmo_class = lambda *args: None

        self.z_inf = z_inf

    # 1. Object properties

    @property
    def mdelta(self):
        """Mass of cluster"""
        return self._get_mass()

    @property
    def cdelta(self):
        """Concentration of cluster"""
        return self._get_concentration()

    @property
    def massdef(self):
        """Definition for the mass of cluster"""
        return self.__massdef

    @property
    def delta_mdef(self):
        """Number of deltas in mass definition of cluster"""
        return self.__delta_mdef

    @property
    def halo_profile_model(self):
        """Halo profile model"""
        return self.__halo_profile_model

    # 1.a Object properties setter

    @mdelta.setter
    def mdelta(self, mdelta):
        """Set mass of cluster"""
        self.set_mass(mdelta)

    @cdelta.setter
    def cdelta(self, cdelta):
        """Set concentration of cluster"""
        self.set_concentration(cdelta)

    @massdef.setter
    def massdef(self, massdef):
        """Set definition for the mass of cluster"""
        self.set_halo_density_profile(
            halo_profile_model=self.halo_profile_model,
            massdef=massdef,
            delta_mdef=self.delta_mdef,
        )

    @delta_mdef.setter
    def delta_mdef(self, delta_mdef):
        """Set number of deltas in mass definition of cluster"""
        self.set_halo_density_profile(
            halo_profile_model=self.halo_profile_model,
            massdef=self.massdef,
            delta_mdef=delta_mdef,
        )

    @halo_profile_model.setter
    def halo_profile_model(self, halo_profile_model):
        """Set halo profile model"""
        self.set_halo_density_profile(
            halo_profile_model=halo_profile_model,
            massdef=self.massdef,
            delta_mdef=self.delta_mdef,
        )

    # 2. Functions to be implemented by children classes

    def _get_mass(self):
        r"""Gets the value of the :math:`M_\Delta`"""
        raise NotImplementedError

    def _get_concentration(self):
        r"""Gets the value of the concentration"""
        raise NotImplementedError

    def _set_mass(self, mdelta):
        r"""Actually sets the value of the :math:`M_\Delta` (without value check)"""
        raise NotImplementedError

    def _set_concentration(self, cdelta):
        r"""Actuall sets the value of the concentration (without value check)"""
        raise NotImplementedError

    def _update_halo_density_profile(self):
        raise NotImplementedError

    def _set_einasto_alpha(self, alpha):
        r"""Actually sets the value of the :math:`\alpha` parameter for the Einasto profile"""
        raise NotImplementedError

    def _get_einasto_alpha(self, z_cl=None):
        r"""Returns the value of the :math:`\alpha` parameter for the Einasto profile,
        if defined"""
        raise NotImplementedError

    def _get_delta_mdef_virial(self, z_cl):
        "Gets the overdensity delta value for massdef='virial' from the backends"
        raise NotImplementedError

    def _set_projected_quad(self, use_projected_quad):
        """Implemented for the CCL backend only"""
        raise NotImplementedError

    def _eval_3d_density(self, r3d, z_cl):
        raise NotImplementedError

    def _eval_surface_density(self, r_proj, z_cl):
        raise NotImplementedError

    def _eval_mean_surface_density(self, r_proj, z_cl):
        raise NotImplementedError

    def _eval_excess_surface_density(self, r_proj, z_cl):
        raise NotImplementedError

    # 3. Functions that can be used by all subclasses

    def _set_cosmo(self, cosmo):
        r"""Sets the cosmology to the internal cosmology object"""
        self.cosmo = cosmo if cosmo is not None else self.cosmo_class()

    def _eval_2halo_term_generic(
        self,
        sph_harm_ord,
        r_proj,
        z_cl,
        halobias=1.0,
        logkbounds=(-5, 5),
        ksteps=1000,
        loglbounds=(0, 6),
        lsteps=500,
    ):
        """eval excess surface density from the 2-halo term"""
        # pylint: disable=protected-access
        da = self.cosmo.eval_da(z_cl)
        rho_m = self.cosmo._get_rho_m(z_cl)

        k_values = np.logspace(logkbounds[0], logkbounds[1], ksteps)
        pk_values = self.cosmo._eval_linear_matter_powerspectrum(k_values, z_cl)
        interp_pk = splrep(k_values, pk_values)
        theta = r_proj / da

        # calculate integral, units [Mpc]**-3
        def __integrand__(l_value, theta):
            k_value = l_value / ((1 + z_cl) * da)
            return l_value * jv(sph_harm_ord, l_value * theta) * splev(k_value, interp_pk)

        l_values = np.logspace(loglbounds[0], loglbounds[1], lsteps)
        kernel = np.array([simpson(__integrand__(l_values, t), x=l_values) for t in theta])
        return halobias * kernel * rho_m / (2 * np.pi * (1 + z_cl) ** 3 * da**2)

    def _eval_surface_density_2h(
        self,
        r_proj,
        z_cl,
        halobias=1.0,
        logkbounds=(-5, 5),
        ksteps=1000,
        loglbounds=(0, 6),
        lsteps=500,
    ):
        """eval surface density from the 2-halo term"""
        return self._eval_2halo_term_generic(
            0, r_proj, z_cl, halobias, logkbounds, ksteps, loglbounds, lsteps
        )

    def _eval_excess_surface_density_2h(
        self,
        r_proj,
        z_cl,
        halobias=1.0,
        logkbounds=(-5, 5),
        ksteps=1000,
        loglbounds=(0, 6),
        lsteps=500,
    ):
        """eval excess surface density from the 2-halo term"""
        return self._eval_2halo_term_generic(
            2, r_proj, z_cl, halobias, logkbounds, ksteps, loglbounds, lsteps
        )

    def _eval_rdelta(self, z_cl):
        delta_mdef = self._get_delta_mdef(z_cl)
        return compute_rdelta(self.mdelta, z_cl, self.cosmo, self.massdef, delta_mdef)

    def _eval_mass_in_radius(self, r3d, z_cl):
        alpha = self._get_einasto_alpha(z_cl) if self.halo_profile_model == "einasto" else None
        delta_mdef = self._get_delta_mdef(z_cl)
        return compute_profile_mass_in_radius(
            r3d,
            z_cl,
            self.cosmo,
            self.mdelta,
            self.cdelta,
            self.massdef,
            delta_mdef,
            self.halo_profile_model,
            alpha,
        )

    def _convert_mass_concentration(
        self, z_cl, massdef=None, delta_mdef=None, halo_profile_model=None, alpha=None
    ):
        alpha1 = self._get_einasto_alpha(z_cl) if self.halo_profile_model == "einasto" else None
        delta_mdef1 = self._get_delta_mdef(z_cl)
        return convert_profile_mass_concentration(
            self.mdelta,
            self.cdelta,
            z_cl,
            self.cosmo,
            massdef=self.massdef,
            delta_mdef=delta_mdef1,
            halo_profile_model=self.halo_profile_model,
            alpha=alpha1,
            massdef2=massdef,
            delta_mdef2=delta_mdef,
            halo_profile_model2=halo_profile_model,
            alpha2=alpha,
        )

    def _get_delta_mdef(self, z_cl):
        if self.massdef == "virial":
            return int(self._get_delta_mdef_virial(z_cl))
        return self.delta_mdef

    # 3.1. Miscentering functions

    def _eval_surface_density_miscentered(self, r_proj, z_cl, r_mis, mis_from_backend):
        # set integrand function
        integrand = None
        if mis_from_backend:
            integrand = self._integrand_surface_density_mis
        elif self.halo_profile_model == "nfw":
            integrand = miscentering.integrand_surface_density_nfw
        elif self.halo_profile_model == "einasto":
            integrand = miscentering.integrand_surface_density_einasto
        elif self.halo_profile_model == "hernquist":
            integrand = miscentering.integrand_surface_density_hernquist
        # get aux arguments
        norm, aux_args = self._miscentering_params(z_cl, mis_from_backend)

        extra_integral = self.halo_profile_model == "einasto" and not mis_from_backend
        return miscentering.integrate_azimuthially_miscentered_surface_density(
            r_proj, r_mis, integrand, norm, aux_args, extra_integral
        )

    def _eval_mean_surface_density_miscentered(self, r_proj, z_cl, r_mis, mis_from_backend):
        # set integrand function
        integrand = None
        if mis_from_backend:
            integrand = self._integrand_mean_surface_density_mis
        elif self.halo_profile_model == "nfw":
            integrand = miscentering.integrand_mean_surface_density_nfw
        elif self.halo_profile_model == "einasto":
            integrand = miscentering.integrand_mean_surface_density_einasto
        elif self.halo_profile_model == "hernquist":
            integrand = miscentering.integrand_mean_surface_density_hernquist
        # get aux arguments
        norm, aux_args = self._miscentering_params(z_cl, mis_from_backend)

        extra_integral = self.halo_profile_model == "einasto" and not mis_from_backend
        return miscentering.integrate_azimuthially_miscentered_mean_surface_density(
            r_proj, r_mis, integrand, norm, aux_args, extra_integral
        )

    def _eval_excess_surface_density_miscentered(self, r_proj, z_cl, r_mis, mis_from_backend):
        return self._eval_mean_surface_density_miscentered(
            r_proj, z_cl, r_mis, mis_from_backend
        ) - self._eval_surface_density_miscentered(r_proj, z_cl, r_mis, mis_from_backend)

    def _integrand_surface_density_mis(self, theta, r_proj, r_mis, z_cl):
        return self.eval_surface_density(
            np.sqrt(r_proj**2.0 + r_mis**2.0 - 2 * r_proj * r_mis * np.cos(theta)), z_cl
        )

    def _integrand_mean_surface_density_mis(self, theta, r_proj, r_mis, z_cl):
        return r_proj * self._integrand_surface_density_mis(theta, r_proj, r_mis, z_cl)

    def _miscentering_params(self, z_cl, mis_from_backend):
        params = None
        if mis_from_backend:
            params = 1, (z_cl,)

        else:
            rho_def = (
                self.cosmo.get_rho_m(z_cl) if self.massdef == "mean" else self.cosmo.get_rho_c(z_cl)
            )
            r_s = self.eval_rdelta(z_cl) / self.cdelta

            delta_mdef = self._get_delta_mdef(z_cl)

            if self.halo_profile_model == "nfw":
                rho_s = (
                    delta_mdef
                    / 3.0
                    * self.cdelta**3.0
                    * rho_def
                    / (np.log(1.0 + self.cdelta) - self.cdelta / (1.0 + self.cdelta))
                )
                params = 2 * r_s * rho_s, (r_s,)

            elif self.halo_profile_model == "einasto":
                alpha_ein = self._get_einasto_alpha(z_cl)
                rho_s = (
                    delta_mdef
                    / 3.0
                    * self.cdelta**3.0
                    * rho_def
                    / (
                        2.0 ** (-3.0 / alpha_ein)
                        * alpha_ein ** (-1.0 + 3.0 / alpha_ein)
                        * np.exp(2.0 / alpha_ein)
                        * gamma(3.0 / alpha_ein)
                        * gammainc(3.0 / alpha_ein, 2.0 / alpha_ein * self.cdelta**alpha_ein)
                    )
                )
                params = 2 * rho_s, (r_s, alpha_ein)

            elif self.halo_profile_model == "hernquist":
                rho_s = (
                    delta_mdef
                    / 3.0
                    * self.cdelta**3.0
                    * rho_def
                    / ((self.cdelta / (1.0 + self.cdelta)) ** 2.0)
                    * 2
                )
                params = r_s * rho_s, (r_s,)

        return params

    # 3.2. All these functions are for the single plane case

    def _eval_tangential_shear_core(self, r_proj, z_cl, z_src):
        delta_sigma = self.eval_excess_surface_density(r_proj, z_cl)
        sigma_c = self.cosmo.eval_sigma_crit(z_cl, z_src)
        return delta_sigma / sigma_c

    def _eval_convergence_core(self, r_proj, z_cl, z_src):
        sigma = self.eval_surface_density(r_proj, z_cl)
        sigma_c = self.cosmo.eval_sigma_crit(z_cl, z_src)
        return sigma / sigma_c

    def _eval_reduced_tangential_shear_core(self, r_proj, z_cl, z_src):
        kappa = self.eval_convergence(r_proj, z_cl, z_src)
        gamma_t = self.eval_tangential_shear(r_proj, z_cl, z_src)
        return compute_reduced_shear_from_convergence(gamma_t, kappa)

    def _eval_magnification_core(self, r_proj, z_cl, z_src):
        kappa = self.eval_convergence(r_proj, z_cl, z_src)
        gamma_t = self.eval_tangential_shear(r_proj, z_cl, z_src)
        return 1.0 / ((1 - kappa) ** 2 - abs(gamma_t) ** 2)

    def _eval_magnification_bias_core(self, r_proj, z_cl, z_src, alpha):
        magnification = self.eval_magnification(r_proj, z_cl, z_src)
        return compute_magnification_bias_from_magnification(magnification, alpha)

    # 4. Wrapper functions for input validation

    def set_mass(self, mdelta):
        r"""Sets the value of the :math:`M_\Delta`.

        Parameters
        ----------
        mdelta : float
            Galaxy cluster mass :math:`M_\Delta` in units of :math:`M_\odot`

        Notes
        -----
            This is equivalent to doing self.mdelta = mdelta
        """
        if self.validate_input:
            validate_argument(locals(), "mdelta", float, argmin=0)
        self._set_mass(mdelta)

    def set_concentration(self, cdelta):
        r"""Sets the concentration

        Parameters
        ----------
        cdelta: float
            Concentration

        Notes
        -----
            This is equivalent to doing self.cdelta = cdelta
        """
        if self.validate_input:
            validate_argument(locals(), "cdelta", float, argmin=0)
        self._set_concentration(cdelta)

    def set_cosmo(self, cosmo):
        r"""Sets the cosmology to the internal cosmology object

        Parameters
        ----------
        cosmo: clmm.Comology object, None
            CLMM Cosmology object. If is None, creates a new instance of self.cosmo_class().
        """
        if self.validate_input:
            if self.cosmo_class() is None:
                raise NotImplementedError
            validate_argument(locals(), "cosmo", self.cosmo_class, none_ok=True)
        self._set_cosmo(cosmo)
        self.cosmo.validate_input = self.validate_input

    def set_halo_density_profile(self, halo_profile_model="nfw", massdef="mean", delta_mdef=200):
        r"""Sets the definitions for the halo profile

        Parameters
        ----------
        halo_profile_model: str
            Halo mass profile, supported options are 'nfw', 'einasto', 'hernquist'
            (letter case independent)
        massdef: str
            Mass definition, supported options are 'mean', 'critical', 'virial'
            (letter case independent)
        delta_mdef: int
            Overdensity number. No effect if massdef='virial'.
        """
        # make case independent
        validate_argument(locals(), "massdef", str)
        validate_argument(locals(), "halo_profile_model", str)
        massdef, halo_profile_model = massdef.lower(), halo_profile_model.lower()

        if self.validate_input:
            validate_argument(locals(), "delta_mdef", int, argmin=0)
            if massdef not in self.mdef_dict:
                raise ValueError(
                    f"Halo density profile mass definition {massdef} not currently supported"
                )
            if halo_profile_model not in self.hdpm_dict:
                raise ValueError(
                    f"Halo density profile model {halo_profile_model} not currently supported"
                )
        # Check if we have already an instance of the required object, if not create one
        if (
            (self.hdpm is None)
            or (self.halo_profile_model != halo_profile_model)
            or (self.massdef != massdef)
            or (self.delta_mdef != delta_mdef)
        ):
            # set internal quantities
            self.__halo_profile_model = halo_profile_model
            self.__massdef = massdef
            self.__delta_mdef = delta_mdef
            # set the profile
            self._update_halo_density_profile()

    def set_einasto_alpha(self, alpha):
        r"""Sets the value of the :math:`\alpha` parameter for the Einasto profile

        Parameters
        ----------
        alpha : float, None
            If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
            cosmology-dependent value for the CCL backend.)
        """
        if self.halo_profile_model != "einasto":
            raise NotImplementedError(
                "The Einasto slope cannot be set "
                "for your combination of profile choice "
                "or modeling backend."
            )
        if self.validate_input:
            validate_argument(locals(), "alpha", float, none_ok=True)
        self._set_einasto_alpha(alpha)

    def get_einasto_alpha(self, z_cl=None):
        r"""Returns the value of the :math:`\alpha` parameter for the Einasto profile, if defined

        Parameters
        ----------
        z_cl : float
            Cluster redshift (required for Einasto with the CCL backend, will be ignored for NC)
        """
        if self.halo_profile_model != "einasto":
            raise ValueError(f"Wrong profile model. Current profile = {self.halo_profile_model}")
        return self._get_einasto_alpha(z_cl)

    def set_projected_quad(self, use_projected_quad):
        r"""Control the use of quad_vec to calculate the surface density profile for
        CCL Einasto profile.

        Parameters
        ----------
        use_projected_quad : bool
            Only available for Einasto profile with CCL as the backend. If True, CCL will use
            quad_vec instead of default FFTLog to calculate the surface density profile.
        """
        if self.halo_profile_model != "einasto" or self.backend != "ccl":
            raise NotImplementedError("This option is only available for the CCL Einasto profile.")
        if self.validate_input:
            validate_argument(locals(), "use_projected_quad", bool)
        self._set_projected_quad(use_projected_quad)

    def eval_3d_density(self, r3d, z_cl, verbose=False):
        r"""Retrieve the 3d density :math:`\rho(r)`.

        Parameters
        ----------
        r3d : array_like, float
            Radial position from the cluster center in :math:`M\!pc`.
        z_cl: float
            Redshift of the cluster

        Returns
        -------
        numpy.ndarray, float
            3-dimensional mass density in units of :math:`M_\odot\ Mpc^{-3}`
        """
        if self.validate_input:
            validate_argument(locals(), "r3d", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", "float_array", argmin=0)

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        return self._eval_3d_density(r3d=r3d, z_cl=z_cl)

    def eval_critical_surface_density_eff(self, z_len, pzbins, pzpdf):
        r"""Computes the 'effective critical surface density'

        .. math::
            \langle \Sigma_{\rm crit}^{-1}\rangle^{-1} =
            \left(\int \frac{1}{\Sigma_{\rm crit}(z)} p(z) dz\right)^{-1}

        where :math:`p(z)` is the source photoz probability density function.
        This comes from the maximum likelihood estimator for evaluating a
        :math:`\Delta\Sigma` profile.

        For the standard :math:`\Sigma_{\text{crit}}(z)` definition, use the `eval_sigma_crit`
        method of the CLMM cosmology object.

        Parameters
        ----------
        z_len : float
            Galaxy cluster redshift
        pzbins : array-like
            Bins where the source redshift pdf is defined
        pzpdf : array-like
            Values of the source redshift pdf


        Returns
        -------
        sigma_c : numpy.ndarray, float
            Cosmology-dependent effective critical surface density in units of
            :math:`M_\odot\ Mpc^{-2}`
        """

        if self.validate_input:
            validate_argument(locals(), "z_len", float, argmin=0)

        def inv_sigmac(redshift):
            return 1.0 / self.cosmo.eval_sigma_crit(z_len, redshift)

        return 1.0 / _integ_pzfuncs(pzpdf, pzbins, kernel=inv_sigmac)

    def eval_surface_density(self, r_proj, z_cl, r_mis=None, mis_from_backend=False, verbose=False):
        r"""Computes the surface mass density

        Parameters
        ----------
        r_proj : array_like
            Projected radial position from the cluster center in :math:`M\!pc`
        z_cl: float
            Redshift of the cluster
        r_mis : float, optional
            Projected miscenter distance in :math:`M\!pc`
        mis_from_backend : bool, optional
            If True, use the projected surface density from the backend for miscentering
            calculations. If False, use the (faster) CLMM exact analytical
            implementation instead. (Default: False)
        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends. (Default: False)

        Returns
        -------
        numpy.ndarray, float
            2D projected surface density in units of :math:`M_\odot\ Mpc^{-2}`

        """
        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            if r_mis is not None:
                validate_argument(locals(), "r_mis", float, argmin=0, eqmin=True)

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        if r_mis is not None:
            return self._eval_surface_density_miscentered(
                r_proj=r_proj, z_cl=z_cl, r_mis=r_mis, mis_from_backend=mis_from_backend
            )
        return self._eval_surface_density(r_proj=r_proj, z_cl=z_cl)

    def eval_mean_surface_density(
        self, r_proj, z_cl, r_mis=None, mis_from_backend=False, verbose=False
    ):
        r"""Computes the mean value of surface density inside radius `r_proj`

        Parameters
        ----------
        r_proj : array_like
            Projected radial position from the cluster center in :math:`M\!pc`.
        z_cl: float
            Redshift of the cluster
        r_mis : float, optional
            Projected miscenter distance in :math:`M\!pc`.
        mis_from_backend : bool, optional
            If True, use the projected surface density from the backend for miscentering
            calculations. If False, use the (faster) CLMM exact analytical
            implementation instead. (Default: False)
        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends. (Default: False)

        Returns
        -------
        numpy.ndarray, float
            Mean surface density in units of :math:`M_\odot\ Mpc^{-2}`.
        """
        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            if r_mis is not None:
                validate_argument(locals(), "r_mis", float, argmin=0)

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        if r_mis is not None:
            return self._eval_mean_surface_density_miscentered(
                r_proj=r_proj, z_cl=z_cl, r_mis=r_mis, mis_from_backend=mis_from_backend
            )

        return self._eval_mean_surface_density(r_proj=r_proj, z_cl=z_cl)

    def eval_excess_surface_density(
        self, r_proj, z_cl, r_mis=None, mis_from_backend=False, verbose=False
    ):
        r"""Computes the excess surface density

        Parameters
        ----------
        r_proj : array_like
            Projected radial position from the cluster center in :math:`M\!pc`.
        z_cl: float
            Redshift of the cluster
        r_mis : float, optional
            Projected miscenter distance in :math:`M\!pc`.
        mis_from_backend : bool, optional
            If True, use the projected surface density from the backend for miscentering
            calculations. If False, use the (faster) CLMM exact analytical
            implementation instead. (Default: False)
        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends. (Default: False)

        Returns
        -------
        numpy.ndarray, float
            Excess surface density in units of :math:`M_\odot\ Mpc^{-2}`.
        """
        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            if r_mis is not None:
                validate_argument(locals(), "r_mis", float, argmin=0, eqmin=True)

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        if r_mis is not None:
            return self._eval_excess_surface_density_miscentered(
                r_proj=r_proj, z_cl=z_cl, r_mis=r_mis, mis_from_backend=mis_from_backend
            )
        return self._eval_excess_surface_density(r_proj=r_proj, z_cl=z_cl)

    def eval_excess_surface_density_2h(
        self,
        r_proj,
        z_cl,
        halobias=1.0,
        logkbounds=(-5, 5),
        ksteps=1000,
        loglbounds=(0, 6),
        lsteps=500,
    ):
        r"""Computes the 2-halo term excess surface density (CCL and NC backends only)

        Parameters
        ----------
        r_proj : array_like
            Projected radial position from the cluster center in :math:`M\!pc`.
        z_cl: float
            Redshift of the cluster
        halobias : float, optional
            Value of the halo bias
        logkbounds : tuple(float, float), shape (2,), optional
           Log10 of the upper and lower bounds for the linear matter power spectrum
        ksteps : int, optional
           Number of steps in k-space
        loglbounds : tuple(float, float), shape (2,), optional
           Log10 of the upper and lower bounds for numerical integration
        lsteps: int, optional
            Number of steps for numerical integration

        Returns
        -------
        numpy.ndarray, float
            Excess surface density from the 2-halo term in units of :math:`M_\odot\ Mpc^{-2}`.
        """

        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "halobias", float, argmin=0)
            validate_argument(locals(), "logkbounds", tuple, shape=(2,))
            validate_argument(locals(), "ksteps", int, argmin=1)
            validate_argument(locals(), "loglbounds", tuple, shape=(2,))
            validate_argument(locals(), "lsteps", int, argmin=1)

        if self.backend not in ("ccl", "nc"):
            raise NotImplementedError(
                f"2-halo term not currently supported with the {self.backend} backend. "
                "Use the CCL or NumCosmo backend instead"
            )
        return self._eval_excess_surface_density_2h(
            r_proj, z_cl, halobias, logkbounds, ksteps, loglbounds, lsteps
        )

    def eval_surface_density_2h(
        self,
        r_proj,
        z_cl,
        halobias=1.0,
        logkbounds=(-5, 5),
        ksteps=1000,
        loglbounds=(0, 6),
        lsteps=500,
    ):
        r"""Computes the 2-halo term surface density (CCL and NC backends only)

        Parameters
        ----------
        r_proj : array_like
            Projected radial position from the cluster center in :math:`M\!pc`.
        z_cl: float
            Redshift of the cluster
        halobias : float, optional
           Value of the halo bias
        logkbounds : tuple(float,float), shape (2,), optional
           Log10 of the upper and lower bounds for the linear matter power spectrum
        ksteps : int, optional
           Number of steps in k-space
        loglbounds : tuple(float,float), shape (2,), optional
           Log10 of the upper and lower bounds for numerical integration
        lsteps: int, optional
            Number of steps for numerical integration

        Returns
        -------
        numpy.ndarray, float
            Excess surface density from the 2-halo term in units of :math:`M_\odot\ Mpc^{-2}`.
        """

        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "halobias", float, argmin=0)
            validate_argument(locals(), "logkbounds", tuple, shape=(2,))
            validate_argument(locals(), "ksteps", int, argmin=1)
            validate_argument(locals(), "loglbounds", tuple, shape=(2,))
            validate_argument(locals(), "lsteps", int, argmin=1)

        if self.backend not in ("ccl", "nc"):
            raise NotImplementedError(
                f"2-halo term not currently supported with the {self.backend} backend. "
                "Use the CCL or NumCosmo backend instead"
            )
        return self._eval_surface_density_2h(
            r_proj, z_cl, halobias, logkbounds, ksteps, loglbounds, lsteps
        )

    def eval_tangential_shear(self, r_proj, z_cl, z_src, z_src_info="discrete", verbose=False):
        r"""Computes the tangential shear

        Parameters
        ----------
        r_proj : array_like
            The projected radial positions in :math:`M\!pc`.
        z_cl : float
            Galaxy cluster redshift
        z_src : array_like, float, function
            Information on the background source galaxy redshift(s). Value required depends on
            `z_src_info` (see below).
        z_src_info : str, optional
            Type of redshift information provided by the `z_src` argument.
            The following supported options are:

                * 'discrete' (default) : The redshift of sources is provided by `z_src`.
                  It can be individual redshifts for each source galaxy when `z_src` is an
                  array or all sources are at the same redshift when `z_src` is a float.

                * 'beta' : The averaged lensing efficiency is provided by `z_src`.
                  `z_src` must be a tuple containing
                  ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
                  the lensing efficiency and square of the lensing efficiency averaged over
                  the galaxy redshift distribution repectively.

                    .. math::
                        \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                        {D_{L,\infty}}\right\rangle

                    .. math::
                        \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                        {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends.

        Returns
        -------
        numpy.ndarray, float
            tangential shear
        """

        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "z_src_info", str)
            self._validate_z_src(locals())

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        # function _validate_z_src already safekeeps from this error:
        # pylint: disable=possibly-used-before-assignment

        if z_src_info == "discrete":
            warning_msg = (
                "\nSome source redshifts are lower than the cluster redshift."
                + "\nShear = 0 for those galaxies."
            )
            gammat = compute_for_good_redshifts(
                self._eval_tangential_shear_core,
                z_cl,
                z_src,
                0.0,
                warning_msg,
                "z_cl",
                "z_src",
                r_proj,
            )
        elif z_src_info == "beta":
            beta_s_mean = z_src[0]
            gammat_inf = self._eval_tangential_shear_core(
                r_proj=r_proj, z_cl=z_cl, z_src=self.z_inf
            )
            gammat = beta_s_mean * gammat_inf

        return gammat

    def eval_convergence(self, r_proj, z_cl, z_src, z_src_info="discrete", verbose=False):
        r"""Computes the mass convergence

        .. math::
            \kappa = \frac{\Sigma}{\Sigma_{crit}}

        or

        .. math::
            \kappa = \kappa_\infty \times <\beta_s>

        Parameters
        ----------
        r_proj : array_like
            The projected radial positions in :math:`M\!pc`.
        z_cl : float
            Galaxy cluster redshift
        z_src : array_like, float, function
            Information on the background source galaxy redshift(s). Value required depends on
            `z_src_info` (see below).
        z_src_info : str, optional
            Type of redshift information provided by the `z_src` argument.
            The following supported options are:

                * 'discrete' (default) : The redshift of sources is provided by `z_src`.
                  It can be individual redshifts for each source galaxy when `z_src` is an
                  array or all sources are at the same redshift when `z_src` is a float.

                * 'beta' : The averaged lensing efficiency is provided by `z_src`.
                  `z_src` must be a tuple containing
                  ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
                  the lensing efficiency and square of the lensing efficiency averaged over
                  the galaxy redshift distribution repectively.

                    .. math::
                        \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                        {D_{L,\infty}}\right\rangle

                    .. math::
                        \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                        {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends.

        Returns
        -------
        numpy.ndarray, float
            Mass convergence, kappa.
        """
        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "z_src_info", str)
            self._validate_z_src(locals())

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        # function _validate_z_src already safekeeps from this error:
        # pylint: disable=possibly-used-before-assignment

        if z_src_info == "discrete":
            warning_msg = (
                "\nSome source redshifts are lower than the cluster redshift."
                + "\nConvergence = 0 for those galaxies."
            )
            kappa = compute_for_good_redshifts(
                self._eval_convergence_core,
                z_cl,
                z_src,
                0.0,
                warning_msg,
                "z_cl",
                "z_src",
                r_proj,
            )
        elif z_src_info == "beta":
            beta_s_mean = z_src[0]
            kappa_inf = self._eval_convergence_core(r_proj=r_proj, z_cl=z_cl, z_src=self.z_inf)
            kappa = beta_s_mean * kappa_inf

        return kappa

    def _pdz_weighted_avg(self, core, pdz_func, r_proj, z_cl, integ_kwargs=None):
        r"""Computes function averaged over PDZ

        Parameters
        ----------
        core : function
            Function to be averaged, must take tangential shear and convergence as input.
        pdz_func : function
            Redshift distribution function. Must be a one dimentional function.
        r_proj : array_like
            The projected radial positions in :math:`M\!pc`.
        z_cl : float
            Galaxy cluster redshift
        integ_kwargs: None, dict
            Extra arguments for the redshift integration (when
            `approx=None, z_src_info='distribution'`). Possible keys are:

                * 'zmin' (None, float) : Minimum redshift to be set as the source of the galaxy
                  when performing the sum. (default=None)
                * 'zmax' (float) : Maximum redshift to be set as the source of the galaxy
                  when performing the sum. (default=10.0)
                * 'delta_z_cut' (float) : Redshift cut so that `zmin` = `z_cl` + `delta_z_cut`.
                  `delta_z_cut` is ignored if `z_min` is already provided. (default=0.1)

        Returns
        -------
        array_like
            Function averaged by pdz, with r_proj dimention.
        """

        def tfunc(z, radius):
            return compute_beta_s_func(
                z,
                z_cl,
                self.z_inf,
                self.cosmo,
                self._eval_tangential_shear_core,
                radius,
                z_cl,
                self.z_inf,
            )

        def kfunc(z, radius):
            return compute_beta_s_func(
                z,
                z_cl,
                self.z_inf,
                self.cosmo,
                self._eval_convergence_core,
                radius,
                z_cl,
                self.z_inf,
            )

        def __integrand__(z, radius):
            return pdz_func(z) * core(tfunc(z, radius), kfunc(z, radius))

        _integ_kwargs = {"zmax": 10.0, "delta_z_cut": 0.1}

        _integ_kwargs.update({} if integ_kwargs is None else integ_kwargs)

        zmax = _integ_kwargs["zmax"]
        delta_z_cut = _integ_kwargs["delta_z_cut"]
        zmin = _integ_kwargs.get("zmin", z_cl + delta_z_cut)

        out = np.array([quad(__integrand__, zmin, zmax, (r))[0] for r in r_proj])
        return out / quad(pdz_func, zmin, zmax)[0]

    def eval_reduced_tangential_shear(
        self,
        r_proj,
        z_cl,
        z_src,
        z_src_info="discrete",
        approx=None,
        integ_kwargs=None,
        verbose=False,
    ):
        r"""Computes the reduced tangential shear

        .. math::
            g_t = \frac{\gamma_t}{1-\kappa}

        Parameters
        ----------
        r_proj : array_like
            The projected radial positions in :math:`M\!pc`.
        z_cl : float
            Galaxy cluster redshift
        z_src : array_like, float, function
            Information on the background source galaxy redshift(s). Value required depends on
            `z_src_info` (see below).
        z_src_info : str, optional
            Type of redshift information provided by the `z_src` argument.
            The following supported options are:

                * 'discrete' (default) : The redshift of sources is provided by `z_src`.
                  It can be individual redshifts for each source galaxy when `z_src` is an
                  array or all sources are at the same redshift when `z_src` is a float
                  (Used for `approx=None`).

                * 'distribution' : A redshift distribution function is provided by `z_src`.
                  `z_src` must be a one dimentional function (Used when `approx=None`).

                * 'beta' : The averaged lensing efficiency is provided by `z_src`.
                  `z_src` must be a tuple containing
                  ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
                  the lensing efficiency and square of the lensing efficiency averaged over
                  the galaxy redshift distribution repectively.

                    .. math::
                        \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                        {D_{L,\infty}}\right\rangle

                    .. math::
                        \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                        {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

        approx : str, optional
            Type of computation to be made for reduced tangential shears, options are:

                * None (default): Requires `z_src_info` to be 'discrete' or 'distribution'.
                  If `z_src_info='discrete'`, full computation is made for each
                  `r_proj, z_src` pair individually. If `z_src_info='distribution'`, reduced
                  tangential shear at each value of `r_proj` is calculated as

                  .. math::
                      g_t
                      =\left<\frac{\beta_s\gamma_{\infty}}{1-\beta_s\kappa_{\infty}}\right>
                      =\frac{\int_{z_{min}}^{z_{max}}\frac{\beta_s(z)\gamma_{\infty}}
                      {1-\beta_s(z)\kappa_{\infty}}N(z)\text{d}z}
                      {\int_{z_{min}}^{z_{max}} N(z)\text{d}z}

                * 'order1' : Same approach as in Weighing the Giants - III (equation 6 in
                  Applegate et al. 2014; https://arxiv.org/abs/1208.0605). `z_src_info` must be
                  'beta':

                  .. math::
                      g_t\approx\frac{\left<\beta_s\right>\gamma_{\infty}}
                      {1-\left<\beta_s\right>\kappa_{\infty}}

                * 'order2' : Same approach as in Cluster Mass Calibration at High
                  Redshift (equation 12 in Schrabback et al. 2017;
                  https://arxiv.org/abs/1611.03866).
                  `z_src_info` must be 'beta':

                  .. math::
                      g_t\approx\frac{\left<\beta_s\right>\gamma_{\infty}}
                      {1-\left<\beta_s\right>\kappa_{\infty}}
                      \left(1+\left(\frac{\left<\beta_s^2\right>}
                      {\left<\beta_s\right>^2}-1\right)\left<\beta_s\right>\kappa_{\infty}\right)

        integ_kwargs: None, dict
            Extra arguments for the redshift integration (when
            `approx=None, z_src_info='distribution'`). Possible keys are:

                * 'zmin' (None, float) : Minimum redshift to be set as the source of the galaxy
                  when performing the sum. (default=None)
                * 'zmax' (float) : Maximum redshift to be set as the source of the galaxy
                  when performing the sum. (default=10.0)
                * 'delta_z_cut' (float) : Redshift cut so that `zmin` = `z_cl` + `delta_z_cut`.
                  `delta_z_cut` is ignored if `z_min` is already provided. (default=0.1)

        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends.

        Returns
        -------
        gt : numpy.ndarray, float
            Reduced tangential shear

        Notes
        -----
        Need to figure out if we want to raise exceptions rather than errors here?
        """
        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "z_src_info", str)
            validate_argument(locals(), "approx", str, none_ok=True)
            self._validate_approx_z_src_info(locals())
            self._validate_z_src(locals())

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        # functions _validate_z_src, _validate_approx_z_src_info already safekeeps from this error:
        # pylint: disable=possibly-used-before-assignment

        if approx is None:
            if z_src_info == "distribution":
                gt = self._pdz_weighted_avg(
                    lambda gammat, kappa: gammat / (1 - kappa),
                    z_src,
                    r_proj,
                    z_cl,
                    integ_kwargs=integ_kwargs,
                )
            elif z_src_info == "discrete":
                warning_msg = (
                    "\nSome source redshifts are lower than the cluster redshift."
                    + "\nReduced_shear = 0 for those galaxies."
                )
                gt = compute_for_good_redshifts(
                    self._eval_reduced_tangential_shear_core,
                    z_cl,
                    z_src,
                    0.0,
                    warning_msg,
                    "z_cl",
                    "z_src",
                    r_proj,
                )
        elif approx in ("order1", "order2"):
            beta_s_mean = z_src[0]

            gammat_inf = self._eval_tangential_shear_core(r_proj, z_cl, z_src=self.z_inf)
            kappa_inf = self._eval_convergence_core(r_proj, z_cl, z_src=self.z_inf)

            gt = beta_s_mean * gammat_inf / (1.0 - beta_s_mean * kappa_inf)

            if approx == "order2":
                beta_s_square_mean = z_src[1]
                gt *= (
                    1.0
                    + (beta_s_square_mean / (beta_s_mean * beta_s_mean) - 1.0)
                    * beta_s_mean
                    * kappa_inf
                )

        return gt

    def eval_magnification(
        self,
        r_proj,
        z_cl,
        z_src,
        z_src_info="discrete",
        approx=None,
        verbose=False,
        integ_kwargs=None,
    ):
        r"""Computes the magnification

        .. math::
            \mu = \frac{1}{(1-\kappa)^2-|\gamma_t|^2}

        Parameters
        ----------
        r_proj : array_like
            The projected radial positions in :math:`M\!pc`.
        z_cl : float
            Galaxy cluster redshift
        z_src : array_like, float, function
            Information on the background source galaxy redshift(s). Value required depends on
            `z_src_info` (see below).
        z_src_info : str, optional
            Type of redshift information provided by the `z_src` argument.
            The following supported options are:

                * 'discrete' (default) : The redshift of sources is provided by `z_src`.
                  It can be individual redshifts for each source galaxy when `z_src` is an
                  array or all sources are at the same redshift when `z_src` is a float
                  (Used for `approx=None`).

                * 'distribution' : A redshift distribution function is provided by `z_src`.
                  `z_src` must be a one dimentional function (Used when `approx=None`).

                * 'beta' : The averaged lensing efficiency is provided by `z_src`.
                  `z_src` must be a tuple containing
                  ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
                  the lensing efficiency and square of the lensing efficiency averaged over
                  the galaxy redshift distribution repectively.

                    .. math::
                        \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                        {D_{L,\infty}}\right\rangle

                    .. math::
                        \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                        {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

        approx : str, optional
            Type of computation to be made for magnifications, options are:

                * None (default): Requires `z_src_info` to be 'discrete' or 'distribution'.
                  If `z_src_info='discrete'`, full computation is made for each
                  `r_proj, z_src` pair individually. If `z_src_info='distribution'`, magnification
                  at each value of `r_proj` is calculated as

                  .. math::
                      \mu
                      =\left<\frac{1}{\left(1-\beta_s\kappa_{\infty}\right)^2
                      -\left(\beta_s\gamma_{\infty}\right)^2}\right>
                      =\frac{\int_{z_{min}}^{z_{max}}\frac{N(z)\text{d}z}
                      {\left(1-\beta_s(z)\kappa_{\infty}\right)^2
                      -\left(\beta_s(z)\gamma_{\infty}\right)^2}}
                      {\int_{z_{min}}^{z_{max}} N(z)\text{d}z}

                * 'order1' : Uses the weak lensing approximation of the magnification with up to
                  first-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
                  (`z_src_info` must be 'beta'):

                  .. math::
                      \mu \approx 1 + 2 \left<\beta_s\right>\kappa_{\infty}

                * 'order2' : Uses the weak lensing approximation of the magnification with up to
                  second-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
                  (`z_src_info` must be 'beta'):

                  .. math::
                      \mu \approx 1 + 2 \left<\beta_s\right>\kappa_{\infty}
                      + 3 \left<\beta_s^2\right>\kappa_{\infty}^2
                      + \left<\beta_s^2\right>\gamma_{\infty}^2

        integ_kwargs: None, dict
            Extra arguments for the redshift integration (when
            `approx=None, z_src_info='distribution'`). Possible keys are:

                * 'zmin' (None, float) : Minimum redshift to be set as the source of the galaxy
                  when performing the sum. (default=None)
                * 'zmax' (float) : Maximum redshift to be set as the source of the galaxy
                  when performing the sum. (default=10.0)
                * 'delta_z_cut' (float) : Redshift cut so that `zmin` = `z_cl` + `delta_z_cut`.
                  `delta_z_cut` is ignored if `z_min` is already provided. (default=0.1)

        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends.

        Returns
        -------
        mu : numpy.ndarray, float
            magnification, mu.

        """
        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "z_src_info", str)
            validate_argument(locals(), "approx", str, none_ok=True)
            self._validate_approx_z_src_info(locals())
            self._validate_z_src(locals())

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        # functions _validate_z_src, _validate_approx_z_src_info already safekeeps from this error:
        # pylint: disable=possibly-used-before-assignment

        if approx is None:
            if z_src_info == "distribution":
                mu = self._pdz_weighted_avg(
                    lambda gammat, kappa: 1 / ((1 - kappa) ** 2 - gammat**2),
                    z_src,
                    r_proj,
                    z_cl,
                    integ_kwargs=integ_kwargs,
                )
            elif z_src_info == "discrete":
                warning_msg = (
                    "\nSome source redshifts are lower than the cluster redshift."
                    + "\nMagnification = 1 for those galaxies."
                )
                mu = compute_for_good_redshifts(
                    self._eval_magnification_core,
                    z_cl,
                    z_src,
                    1.0,
                    warning_msg,
                    "z_cl",
                    "z_src",
                    r_proj,
                )
        elif approx in ("order1", "order2"):
            beta_s_mean = z_src[0]

            kappa_inf = self._eval_convergence_core(r_proj, z_cl, z_src=self.z_inf)
            gammat_inf = self._eval_tangential_shear_core(r_proj, z_cl, z_src=self.z_inf)

            mu = 1 + 2 * beta_s_mean * kappa_inf

            if approx == "order2":
                beta_s_square_mean = z_src[1]
                # Taylor expansion with up to second-order terms
                mu += 3 * beta_s_square_mean * kappa_inf**2 + beta_s_square_mean * gammat_inf**2

        return mu

    def eval_magnification_bias(
        self,
        r_proj,
        z_cl,
        z_src,
        alpha,
        z_src_info="discrete",
        approx=None,
        integ_kwargs=None,
        verbose=False,
    ):
        r"""Computes the magnification bias

        .. math::
            \mu^{\alpha - 1}

        Parameters
        ----------
        r_proj : array_like
            The projected radial positions in :math:`M\!pc`.
        z_cl : float
            Galaxy cluster redshift
        z_src : array_like, float, function
            Information on the background source galaxy redshift(s). Value required depends on
            `z_src_info` (see below).
        alpha : float
            Slope of the cummulative number count of background sources at a given magnitude
        z_src_info : str, optional
            Type of redshift information provided by the `z_src` argument.
            The following supported options are:

                * 'discrete' (default) : The redshift of sources is provided by `z_src`.
                  It can be individual redshifts for each source galaxy when `z_src` is an
                  array or all sources are at the same redshift when `z_src` is a float
                  (Used for `approx=None`).

                * 'distribution' : A redshift distribution function is provided by `z_src`.
                  `z_src` must be a one dimentional function (Used when `approx=None`).

                * 'beta' : The averaged lensing efficiency is provided by `z_src`.
                  `z_src` must be a tuple containing
                  ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
                  the lensing efficiency and square of the lensing efficiency averaged over
                  the galaxy redshift distribution repectively.

                    .. math::
                        \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                        {D_{L,\infty}}\right\rangle

                    .. math::
                        \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                        {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

        approx : str, optional
            Type of computation to be made for magnification biases, options are:

                * None (default): Requires `z_src_info` to be 'discrete' or 'distribution'.
                  If `z_src_info='discrete'`, full computation is made for each
                  `r_proj, z_src` pair individually. If `z_src_info='distribution'`, magnification
                  bias at each value of `r_proj` is calculated as

                  .. math::
                      \mu^{\alpha-1}
                      &=\left(\left<\frac{1}{\left(1-\beta_s\kappa_{\infty}\right)^2
                      -\left(\beta_s\gamma_{\infty}\right)^2}\right>\right)^{\alpha-1}
                      \\\\
                      &=\frac{\int_{z_{min}}^{z_{max}}\frac{N(z)\text{d}z}
                      {\left(\left(1-\beta_s(z)\kappa_{\infty}\right)^2
                      -\left(\beta_s(z)\gamma_{\infty}\right)^2\right)^{\alpha-1}}}
                      {\int_{z_{min}}^{z_{max}} N(z)\text{d}z}

                * 'order1' : Uses the weak lensing approximation of the magnification bias with up
                  to first-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
                  (`z_src_info` must be 'beta'):

                  .. math::
                      \mu^{\alpha-1} \approx
                      1 + \left(\alpha-1\right)\left(2 \left<\beta_s\right>\kappa_{\infty}\right)

                * 'order2' : Uses the weak lensing approximation of the magnification bias with up
                  to second-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
                  (`z_src_info` must be 'beta'):

                  .. math::
                      \mu^{\alpha-1} \approx
                      1 &+ \left(\alpha-1\right)\left(2 \left<\beta_s\right>\kappa_{\infty}\right)
                      \\\\
                      &+ \left(\alpha-1\right)\left(\left<\beta_s^2\right>\gamma_{\infty}^2\right)
                      \\\\
                      &+ \left(2\alpha-1\right)\left(\alpha-1\right)
                      \left(\left<\beta_s^2\right>\kappa_{\infty}^2\right)

        integ_kwargs: None, dict
            Extra arguments for the redshift integration (when
            `approx=None, z_src_info='distribution'`). Possible keys are:

                * 'zmin' (None, float) : Minimum redshift to be set as the source of the galaxy
                  when performing the sum. (default=None)
                * 'zmax' (float) : Maximum redshift to be set as the source of the galaxy
                  when performing the sum. (default=10.0)
                * 'delta_z_cut' (float) : Redshift cut so that `zmin` = `z_cl` + `delta_z_cut`.
                  `delta_z_cut` is ignored if `z_min` is already provided. (default=0.1)

        verbose : bool, optional
            If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and
            CCL backends.

        Returns
        -------
        mu_bias : numpy.ndarray, float
            magnification bias.

        """
        if self.validate_input:
            validate_argument(locals(), "r_proj", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "z_src_info", str)
            validate_argument(locals(), "alpha", "float_array")
            validate_argument(locals(), "approx", str, none_ok=True)
            self._validate_approx_z_src_info(locals())
            self._validate_z_src(locals())

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha = {self._get_einasto_alpha(z_cl=z_cl)}")

        # functions _validate_z_src, _validate_approx_z_src_info already safekeeps from this error:
        # pylint: disable=possibly-used-before-assignment

        if approx is None:
            # z_src (float or array) is redshift
            if z_src_info == "distribution":
                mu_bias = self._pdz_weighted_avg(
                    lambda gammat, kappa: 1 / ((1 - kappa) ** 2 - gammat**2) ** (alpha - 1),
                    z_src,
                    r_proj,
                    z_cl,
                    integ_kwargs=integ_kwargs,
                )
            elif z_src_info == "discrete":
                warning_msg = (
                    "\nSome source redshifts are lower than the cluster redshift."
                    + "\nMagnification bias = 1 for those galaxies."
                )
                mu_bias = compute_for_good_redshifts(
                    self._eval_magnification_bias_core,
                    z_cl,
                    z_src,
                    1.0,
                    warning_msg,
                    "z_cl",
                    "z_src",
                    r_proj,
                    alpha=alpha,
                )

        elif approx in ("order1", "order2"):
            beta_s_mean = z_src[0]

            kappa_inf = self._eval_convergence_core(r_proj, z_cl, z_src=self.z_inf)
            gammat_inf = self._eval_tangential_shear_core(r_proj, z_cl, z_src=self.z_inf)

            mu_bias = 1 + (alpha - 1) * (2 * beta_s_mean * kappa_inf)

            if approx == "order2":
                beta_s_square_mean = z_src[1]
                # Taylor expansion with up to second-order terms
                mu_bias += (alpha - 1) * (beta_s_square_mean * gammat_inf**2) + (
                    2 * alpha - 1
                ) * (alpha - 1) * beta_s_square_mean * kappa_inf**2

        return mu_bias

    def eval_rdelta(self, z_cl):
        r"""Retrieves the radius for mdelta

        .. math::
            r_\Delta=\left(\frac{3 M_\Delta}{4 \pi \Delta \rho_{bkg}(z)}\right)^{1/3}

        Parameters
        ----------
        z_cl: float
            Redshift of the cluster

        Returns
        -------
        float
            Radius in :math:`M\!pc`.
        """
        if self.validate_input:
            validate_argument(locals(), "z_cl", float, argmin=0)

        return self._eval_rdelta(z_cl)

    def eval_mass_in_radius(self, r3d, z_cl, verbose=False):
        r"""Computes the mass inside a given radius of the profile.
        The mass is calculated as

        .. math::
            M(<\text{r3d}) = M_{\Delta}\;
            \frac{f\left(\frac{\text{r3d}}{r_{\Delta}/c_{\Delta}}\right)}{f(c_{\Delta})},

        where :math:`f(x)` for the different models are

        NFW:

        .. math::
            \quad \ln(1+x)-\frac{x}{1+x}

        Einasto: (:math:`\gamma` is the lower incomplete gamma function)

        .. math::
            \gamma(\frac{3}{\alpha}, \frac{2}{\alpha}x^{\alpha})

        Hernquist:

        .. math::
            \left(\frac{x}{1+x}\right)^2

        Parameters
        ----------
        r3d : array_like, float
            Radial position from the cluster center in :math:`M\!pc`.
        z_cl: float
            Redshift of the cluster

        Returns
        -------
        numpy.ndarray, float
            Mass in units of :math:`M_\odot`
        """
        if self.validate_input:
            validate_argument(locals(), "r3d", "float_array", argmin=0)
            validate_argument(locals(), "z_cl", float, argmin=0)

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha (in) = {self._get_einasto_alpha(z_cl=z_cl)}")

        return self._eval_mass_in_radius(r3d, z_cl)

    def convert_mass_concentration(
        self,
        z_cl,
        massdef=None,
        delta_mdef=None,
        halo_profile_model=None,
        alpha=None,
        verbose=False,
    ):
        r"""Converts current mass and concentration to the values for a different model.

        Parameters
        ----------
        z_cl: float
            Redshift of the cluster
        massdef : str, None
            Profile mass definition to convert to ("mean", "critical", "virial").
            If None, same value of current model is used.
        delta_mdef : int, None
            Mass overdensity definition to convert to.
            If None, same value of current model is used.
        halo_profile_model : str, None
            Profile model parameterization to convert to ("nfw", "einasto", "hernquist").
            If None, same value of current model is used.
        alpha : float, None
            Einasto slope to convert to when `halo_profile_model='einasto'`.
            If None, same value of current model is used.

        Returns
        -------
        float
            Mass of different model in units of :math:`M_\odot`.
        float
            Concentration of different model.
        """
        if self.validate_input:
            validate_argument(locals(), "z_cl", float, argmin=0)
            validate_argument(locals(), "massdef", str, none_ok=True)
            validate_argument(locals(), "delta_mdef", int, argmin=0, none_ok=True)
            validate_argument(locals(), "halo_profile_model", str, none_ok=True)
            validate_argument(locals(), "alpha", "float_array", none_ok=True)

        if self.halo_profile_model == "einasto" and verbose:
            print(f"Einasto alpha (in) = {self._get_einasto_alpha(z_cl=z_cl)}")

        if (
            halo_profile_model == "einasto"
            or (self.halo_profile_model == "einasto" and halo_profile_model is None)
        ) and verbose:
            print(
                "Einasto alpha (out) = "
                f"{self._get_einasto_alpha(z_cl=z_cl) if alpha is None else alpha}"
            )

        return self._convert_mass_concentration(
            z_cl, massdef, delta_mdef, halo_profile_model, alpha
        )

    def _validate_z_src(self, loc_dict):
        r"""Validation for z_src according to z_src_info. The conditions are:

            * z_src_info='discrete' : z_src must be array or float.
            * z_src_info='distribution' : z_src must be a one dimentional function.
            * z_src_info='beta' : z_src must be a tuple containing
              ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`).

        Parameters
        ----------
        locals_dict: dict
            Should be the call locals()
        """
        if loc_dict["z_src_info"] == "discrete":
            validate_argument(loc_dict, "z_src", "float_array", argmin=0)
        elif loc_dict["z_src_info"] == "distribution":
            validate_argument(loc_dict, "z_src", "function", none_ok=False)
            integ_kwargs = {} if loc_dict["integ_kwargs"] is None else loc_dict["integ_kwargs"]
            _def_keys = ["zmin", "zmax", "delta_z_cut"]
            if any(key not in _def_keys for key in integ_kwargs):
                raise KeyError(
                    f"integ_kwargs must contain only {_def_keys} keys, "
                    f" {integ_kwargs.keys()} provided."
                )
        elif loc_dict["z_src_info"] == "beta":
            validate_argument(loc_dict, "z_src", "array")
            beta_info = {
                "beta_s_mean": loc_dict["z_src"][0],
                "beta_s_square_mean": loc_dict["z_src"][1],
            }
            validate_argument(beta_info, "beta_s_mean", "float_array")
            validate_argument(beta_info, "beta_s_square_mean", "float_array")
        else:
            raise ValueError(f"Unsupported z_src_info (='{loc_dict['z_src_info']}')")

    def _validate_approx_z_src_info(self, loc_dict):
        r"""Validation for compatility between approx and z_src_info. The conditions are:

            * approx=None: z_src_info must be 'discrete' or 'distribution'
            * approx='order1' or 'order2': z_src_info must be 'beta'
            * approx=other: raises error

        Parameters
        ----------
        locals_dict: dict
            Should be the call locals()
        """
        # check compatility between approx and z_src_info
        z_src_info, approx = loc_dict["z_src_info"], loc_dict["approx"]
        if approx is None:
            if z_src_info not in ("discrete", "distribution"):
                raise ValueError(
                    "approx=None requires z_src_info='discrete' or 'distribution',"
                    f" z_src_info='{z_src_info}' was provided."
                )
        elif approx in ("order1", "order2"):
            if z_src_info != "beta":
                raise ValueError(
                    f"approx='{approx}' requires z_src_info='beta', "
                    f"z_src_info='{z_src_info}' was provided."
                )
        else:
            raise ValueError(f"Unsupported approx (='{approx}')")



"""@file clusterensemble.py
The Cluster Ensemble class
"""

import pickle
import warnings

import healpy
import numpy as np

from .dataops import make_stacked_radial_profile
from .gcdata import GCData
from .utils import DiffArray


class ClusterEnsemble:
    """Object that contains a list of GalaxyCluster objects

    Attributes
    ----------
    unique_id : int or string
        Unique identifier of the galaxy cluster ensemble
    data : GCData
        Table with galaxy clusters data (i. e. ids, profiles, redshifts).
    id_dict: dict
        Dictionary of indices given the cluster id
    stacked_data : GCData, None
        Stacked cluster profiles
    cov : dict
        Dictionary with the covariances:

        * "tan_sc" : tangential component computed with sample covariance
        * "cross_sc" : cross component computed with sample covariance
        * "tan_bs" : tangential component computed with bootstrap
        * "cross_bs" : cross component computed with bootstrap
        * "tan_jk" : tangential component computed with jackknife
        * "cross_jk" : cross component computed with jackknife

    """

    def __init__(self, unique_id, gc_list=None, **kwargs):
        """Initializes a ClusterEnsemble object
        Parameters
        ----------
        unique_id : int or string
            Unique identifier of the galaxy cluster ensemble
        """
        if isinstance(unique_id, (int, str)):
            unique_id = str(unique_id)
        else:
            raise TypeError(f"unique_id incorrect type: {type(unique_id)}")
        self.unique_id = unique_id
        self.data = GCData(meta={"bin_units": None, "radius_min": None, "radius_max": None})
        if gc_list is not None:
            self._add_values(gc_list, **kwargs)
            weights_out = kwargs.get("weights_out", "W_l")
            if (self.data[weights_out] == 0).any():
                warnings.warn(
                    "There are empty bins in some of the profile tables,"
                    f" filter them with {weights_out}>0 for visualization."
                )
        self.stacked_data = None
        self.cov = {
            "tan_sc": None,
            "cross_sc": None,
            "tan_bs": None,
            "cross_bs": None,
            "tan_jk": None,
            "cross_jk": None,
        }

    def _add_values(self, gc_list, **kwargs):
        """Add values for all attributes

        Parameters
        ----------
        gc_list : list, tuple
            List of GalaxyCluster objects.
        gc_cols : list, tuple
            List of GalaxyCluster objects.
        """
        for cluster in gc_list:
            self.make_individual_radial_profile(cluster, **kwargs)

    def __getitem__(self, item):
        """Returns self.data[item]"""
        return self.data[item]

    def __len__(self):
        """Returns length of ClusterEnsemble"""
        return len(self.data)

    def make_individual_radial_profile(
        self,
        galaxycluster,
        bin_units,
        bins=10,
        error_model="ste",
        cosmo=None,
        tan_component_in="et",
        cross_component_in="ex",
        tan_component_out="gt",
        cross_component_out="gx",
        tan_component_in_err=None,
        cross_component_in_err=None,
        use_weights=True,
        weights_in="w_ls",
        weights_out="W_l",
    ):
        """Compute the individual shear profile from a single GalaxyCluster object
        and adds the averaged data in the data attribute.

        Parameters
        ----------
        galaxycluster : GalaxyCluster
            GalaxyCluster object with cluster metadata and background galaxy data
        bin_units : str
            Units to use for the radial bins of the shear profile
            Allowed Options = ["radians", "deg", "arcmin", "arcsec", "kpc", "Mpc"]
            (letter case independent)
        bins : array_like or int, optional
            User defined bins to use for the shear profile. If a list is provided, use that as
            the bin edges. If a integer is provided, create that many equally spaced bins between
            the minimum and maximum angular separations in bin_units. If nothing is provided,
            defaults to 10 equally spaced bins.
        error_model : str, optional
            Statistical error model to use for y uncertainties. (letter case independent)

                * `ste` - Standard error [=std/sqrt(n) in unweighted computation] (Default).
                * `std` - Standard deviation.

        cosmo: clmm.Comology, optional
            CLMM Cosmology object
        tan_component_in: string, optional
            Name of the tangential component column in `galcat` to be binned.
            Default: 'et'
        cross_component_in: string, optional
            Name of the cross component column in `galcat` to be binned.
            Default: 'ex'
        tan_component_out: string, optional
            Name of the tangetial component binned column to be added in profile table.
            Default: 'gt'
        cross_component_out: string, optional
            Name of the cross component binned profile column to be added in profile table.
            Default: 'gx'
        tan_component_in_err: string, None, optional
            Name of the tangential component error column in `galcat` to be binned.
            Default: None
        cross_component_in_err: string, None, optional
            Name of the cross component error column in `galcat` to be binned.
            Default: None
        weights_in : str, None
            Name of the weight column in `galcat` to be considered in binning.
        weights_out : str, None
            Name of the weight column to be used in the added to the profile table.
        """
        # pylint: disable=unused-argument
        tb_kwargs = {}
        tb_kwargs.update(locals())
        tb_kwargs.pop("self")
        tb_kwargs.pop("tb_kwargs")
        tb_kwargs.pop("galaxycluster")

        cl_bin_units = galaxycluster.galcat.meta.get("bin_units", None)
        self.data.update_info_ext_valid("bin_units", self.data, cl_bin_units, overwrite=False)

        cl_cosmo = galaxycluster.galcat.meta.get("cosmo", None)
        self.data.update_info_ext_valid("cosmo", self.data, cl_cosmo, overwrite=False)

        profile_table = galaxycluster.make_radial_profile(
            include_empty_bins=True, gal_ids_in_bins=False, add=False, **tb_kwargs
        )

        self.add_individual_radial_profile(
            galaxycluster,
            profile_table,
            tan_component_out,
            cross_component_out,
            weights_out,
        )

    def add_individual_radial_profile(
        self,
        galaxycluster,
        profile_table,
        tan_component="gt",
        cross_component="gx",
        weights="W_l",
    ):
        """Compute the individual shear profile from a single GalaxyCluster object
        and adds the averaged data in the data attribute.

        Parameters
        ----------
        galaxycluster : GalaxyCluster
            GalaxyCluster object with cluster metadata and background galaxy data
        profile_table : GCData
            Table containing the radius grid points, the tangential, cross shear and weights
            profiles on that grid.
        tan_component: string, optional
            Name of the tangetial component binned column in the profile table.
            Default: 'gt'
        cross_component: string, optional
            Name of the cross component binned profile column in the profile table.
            Default: 'gx'
        weights : str, None
            Name of the weight binned column in the profile table.
        """
        cl_bin_units = profile_table.meta.get("bin_units", None)
        self.data.update_info_ext_valid("bin_units", self.data, cl_bin_units, overwrite=False)
        for col in ("radius_min", "radius_max"):
            value = DiffArray(profile_table[col])
            self.data.update_info_ext_valid(col, self.data, value, overwrite=False)

        cl_cosmo = profile_table.meta.get("cosmo", None)
        self.data.update_info_ext_valid("cosmo", self.data, cl_cosmo, overwrite=False)

        tbcols = ("radius", tan_component, cross_component, weights)
        data_to_save = [
            galaxycluster.unique_id,
            galaxycluster.ra,
            galaxycluster.dec,
            galaxycluster.z,
            *(np.array(profile_table[col]) for col in tbcols),
        ]
        if len(self.data) == 0:
            for col, data in zip(["cluster_id", "ra", "dec", "z", *tbcols], data_to_save):
                self.data[col] = [data]
        else:
            self.data.add_row(data_to_save)

    def _check_empty_data(self):
        if len(self.data) == 0:
            raise ValueError(
                "There is no single cluster profile data. Please run"
                + "'make_individual_radial_profile' or "
                "'add_individual_radial_profile' "
                "for each cluster in your catalog"
            )

    def make_stacked_radial_profile(self, tan_component="gt", cross_component="gx", weights="W_l"):
        """Computes stacked profile and mean separation distances and add it internally
        to `stacked_data`.

        Parameters
        ----------
        tan_component : string, optional
            Name of the tangential component column in `data`.
            Default: 'gt'
        cross_component : string, optional
            Name of the cross component column in `data`.
            Default: 'gx'
        weights : str
            Name of the weights column in `data`.
        """
        self._check_empty_data()

        radius, components = make_stacked_radial_profile(
            self.data["radius"],
            self.data[weights],
            [self.data[tan_component], self.data[cross_component]],
        )
        self.stacked_data = GCData(
            [
                self.data.meta["radius_min"].value,
                self.data.meta["radius_max"].value,
                radius,
                *components,
            ],
            meta={k: v for k, v in self.data.meta.items() if k not in ("radius_min", "radius_max")},
            names=("radius_min", "radius_max", "radius", tan_component, cross_component),
        )

    def compute_sample_covariance(self, tan_component="gt", cross_component="gx"):
        """Compute Sample covariance matrix for cross and tangential and cross
        stacked profiles and updates .cov dict (`tan_sc`, `cross_sc`).

        Parameters
        ----------
        tan_component : string, optional
            Name of the tangential component column in `data`.
            Default: 'gt'
        cross_component : string, optional
            Name of the cross component column in `data`.
            Default: 'gx'
        """
        self._check_empty_data()

        n_catalogs = len(self.data)
        self.cov["tan_sc"] = np.cov(self.data[tan_component].T, bias=False) / n_catalogs
        self.cov["cross_sc"] = np.cov(self.data[cross_component].T, bias=False) / n_catalogs

    def compute_bootstrap_covariance(
        self, tan_component="gt", cross_component="gx", n_bootstrap=10
    ):
        """Compute the bootstrap covariance matrix, add boostrap covariance matrix for
        tangential and cross stacked profiles and updates .cov dict (`tan_jk`, `cross_bs`).

        Parameters
        ----------
        tan_component : string, optional
            Name of the tangential component column in `data`.
            Default: 'gt'
        cross_component : string, optional
            Name of the cross component column in `data`.
            Default: 'gx'
        n_bootstrap : int
            number of bootstrap resamplings
        """
        self._check_empty_data()
        n_catalogs = len(self)

        cluster_index = np.arange(n_catalogs)
        cluster_index_bootstrap = [
            np.random.choice(cluster_index, n_catalogs) for n_boot in range(n_bootstrap)
        ]

        gt_boot, gx_boot = make_stacked_radial_profile(
            self["radius"][None, cluster_index_bootstrap][0].transpose(1, 2, 0),
            self["W_l"][None, cluster_index_bootstrap][0].transpose(1, 2, 0),
            [
                self[tan_component][None, cluster_index_bootstrap][0].transpose(1, 2, 0),
                self[cross_component][None, cluster_index_bootstrap][0].transpose(1, 2, 0),
            ],
        )[1]

        coeff = (n_catalogs / (n_catalogs - 1)) ** 2
        self.cov["tan_bs"] = coeff * np.cov(np.array(gt_boot), bias=False, ddof=0)
        self.cov["cross_bs"] = coeff * np.cov(np.array(gx_boot), bias=False)

    def compute_jackknife_covariance(self, tan_component="gt", cross_component="gx", n_side=16):
        """Compute the jackknife covariance matrix, add boostrap covariance matrix for
        tangential and cross stacked profiles and updates .cov dict (`tan_jk`, `cross_jk`).

        Uses healpix sky area sub-division : https://healpix.sourceforge.io

        Parameters
        ----------
        tan_component : string, optional
            Name of the tangential component column in `data`.
            Default: 'gt'
        cross_component : string, optional
            Name of the cross component column in `data`.
            Default: 'gx'
        n_side : int
            healpix sky area division parameter (number of sky area : 12*n_side^2)
        """
        # may induce artificial noise if there are some healpix pixels
        # not covering entirely the 2D map of clusters
        self._check_empty_data()

        pixels = healpy.ang2pix(n_side, self.data["ra"], self.data["dec"], nest=True, lonlat=True)
        pixels_list_unique = np.unique(pixels)
        gt_jack, gx_jack = [], []
        for hp_list_delete in pixels_list_unique:
            mask = ~np.isin(pixels, hp_list_delete)
            gt, gx = make_stacked_radial_profile(
                self["radius"][mask],
                self["W_l"][mask],
                [self[tan_component][mask], self[cross_component][mask]],
            )[1]
            gt_jack.append(gt)
            gx_jack.append(gx)
        n_jack = pixels_list_unique.size
        coeff = (n_jack - 1) ** 2 / (n_jack)
        self.cov["tan_jk"] = coeff * np.cov(np.transpose(gt_jack), bias=False, ddof=0)
        self.cov["cross_jk"] = coeff * np.cov(np.transpose(gx_jack), bias=False, ddof=0)

    def save(self, filename, **kwargs):
        """Saves GalaxyCluster object to filename using Pickle"""
        with open(filename, "wb") as fin:
            pickle.dump(self, fin, **kwargs)

    @classmethod
    def load(cls, filename, **kwargs):
        """Loads GalaxyCluster object to filename using Pickle"""
        with open(filename, "rb") as fin:
            self = pickle.load(fin, **kwargs)
        return self



"""@file cluster_toolkit.py
Cosmology using AstroPy (for cluster_toolkit)
"""

import numpy as np
from astropy import units
from astropy.cosmology import FlatLambdaCDM, LambdaCDM

from ..utils.constants import Constants as const
from .parent_class import CLMMCosmology

__all__ = []


class AstroPyCosmology(CLMMCosmology):
    """
    Cosmology object

    Attributes
    ----------
    backend: str
        Name of back-end used
    be_cosmo: cosmology library
        Cosmology library used in the back-end
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # this tag will be used to check if the cosmology object is accepted by the modeling
        self.backend = "ct"

        assert isinstance(self.be_cosmo, LambdaCDM)

    def _init_from_cosmo(self, be_cosmo):
        assert isinstance(be_cosmo, LambdaCDM)
        self.be_cosmo = be_cosmo

    def _init_from_params(self, H0, Omega_b0, Omega_dm0, Omega_k0):
        # pylint: disable=arguments-differ
        kwargs = {
            "H0": H0,
            "Om0": Omega_b0 + Omega_dm0,
            "Ob0": Omega_b0,
            "Tcmb0": 2.7255,
            "Neff": 3.046,
            "m_nu": ([0.06, 0.0, 0.0] * units.eV),
        }
        self.be_cosmo = FlatLambdaCDM(**kwargs)
        if Omega_k0 != 0.0:
            kwargs["Ode0"] = self.be_cosmo.Ode0 - Omega_k0
            self.be_cosmo = LambdaCDM(**kwargs)

    def _set_param(self, key, value):
        raise NotImplementedError("Astropy do not support changing parameters")

    def _get_param(self, key):
        if key == "Omega_m0":
            value = self.be_cosmo.Om0
        elif key == "Omega_b0":
            value = self.be_cosmo.Ob0
        elif key == "Omega_dm0":
            value = self.be_cosmo.Odm0
        elif key == "Omega_k0":
            value = self.be_cosmo.Ok0
        elif key == "h":
            value = self.be_cosmo.H0.to_value() / 100.0
        elif key == "H0":
            value = self.be_cosmo.H0.to_value()
        else:
            raise ValueError(f"Unsupported parameter {key}")
        return value

    def _get_Omega_m(self, z):
        return self.be_cosmo.Om(z)

    def _get_E2(self, z):
        return (self.be_cosmo.H(z) / self.be_cosmo.H0) ** 2

    def _get_E2Omega_m(self, z):
        return self.be_cosmo.Om(z) * (self.be_cosmo.H(z) / self.be_cosmo.H0) ** 2

    def _get_rho_c(self, z):
        return self.be_cosmo.critical_density(z).to(units.Msun / units.Mpc**3).value

    def _eval_da_z1z2_core(self, z1, z2):
        return self.be_cosmo.angular_diameter_distance_z1z2(z1, z2).to_value(units.Mpc)

    def _eval_sigma_crit_core(self, z_len, z_src):
        # Constants
        clight_pc_s = const.CLIGHT_KMS.value * 1000.0 / const.PC_TO_METER.value
        gnewt_pc3_msun_s2 = const.GNEWT.value * const.SOLAR_MASS.value / const.PC_TO_METER.value**3

        d_l = self._eval_da_z1z2_core(0, z_len)
        d_s = self._eval_da_z1z2_core(0, z_src)
        d_ls = self._eval_da_z1z2_core(z_len, z_src)

        return clight_pc_s**2 / (4.0 * np.pi * gnewt_pc3_msun_s2) * d_s / (d_l * d_ls) * 1.0e6



"""Redshift distribution functions"""

import numpy as np
from scipy.special import gamma, gammainc


def _functional_form(redshift, alpha, beta, redshift0, is_cdf=False):
    r"""
    A private function that returns the functionnal form of the redshift distribution used in
    Chang et al (2013):

    .. math::
       P(z) = z^{\\alpha}\times\exp^{\left(-\frac{z}{z0}^\beta\right)}

    Parameters
    ----------
    redshift : float
        Galaxy redshift
    alpha, beta, z0 : floats
        Parameters describing the function
    is_cdf : bool
        If True, returns cumulative function.

    Returns
    -------
    The value of the function at z
    """
    if is_cdf:
        return (
            redshift0 ** (alpha + 1)
            * gammainc((alpha + 1) / beta, (redshift / redshift0) ** beta)
            / beta
            * gamma((alpha + 1) / beta)
        )
    return (redshift**alpha) * np.exp(-((redshift / redshift0) ** beta))


def chang2013(redshift, is_cdf=False):
    """
    Chang et al (2013) unnormalized galaxy redshift distribution function, with the fiducial
    set of parameters.

    Parameters
    ----------
    redshift : float
        Galaxy redshift
    is_cdf : bool
        If True, returns cumulative distribution function.

    Returns
    -------
    The value of the distribution at z
    """
    alpha, beta, redshift0 = 1.24, 1.01, 0.51
    return _functional_form(redshift, alpha, beta, redshift0, is_cdf)


def desc_srd(redshift, is_cdf=False):
    """
    Unnormalized galaxy redshift distribution function used in
    the LSST/DESC Science Requirement Document (arxiv:1809.01669).

    Parameters
    ----------
    redshift : float
        Galaxy redshift
    is_cdf : bool
        If True, returns cumulative distribution function.

    Returns
    -------
    The value of the distribution at z
    """
    alpha, beta, redshift0 = 2.0, 0.9, 0.28
    return _functional_form(redshift, alpha, beta, redshift0, is_cdf)



"""General utility functions that are used in multiple modules"""

import numpy as np


def compute_nfw_boost(r_proj, r_scale, boost0=0.1):
    r"""Computes the boost factor at radii r_proj using a parametric form
    following that of the analytical NFW excess surface density.

    Setting :math:`x = r_{\rm proj}/r_{\rm scale}`

    .. math::
        B(x) = 1 + B_0\frac{1-F(x)}{(x)^2 -1}

    where

    .. math::
        F(x) = \begin{cases} \frac{\arctan \sqrt{x^2 -1 }}{\sqrt{x^2 -1 }} & \text{if $x>1$}, \\
        \text{1} & \text{if x = 1}, \\
        \frac{\text{arctanh} \sqrt{1-x^2 }}{\sqrt{1- x^2 }} & \text{if $x<1$}.
        \end{cases}


    Parameters
    ----------
    r_proj : array_like
        Radii
    r_scale : float
        Scale radius in same units as r_proj
    boost0: float, optional
        Boost factor normalisation

    Returns
    -------
    array
        Boost factor at each value of r_proj
    """

    r_norm = np.array(r_proj) / r_scale

    def _calc_finternal(r_norm):
        radicand = r_norm**2 - 1

        finternal = (
            -1j
            * np.log(
                (1 + np.lib.scimath.sqrt(radicand) * 1j) / (1 - np.lib.scimath.sqrt(radicand) * 1j)
            )
            / (2 * np.lib.scimath.sqrt(radicand))
        )

        return np.nan_to_num((1 - finternal) / radicand, nan=1.0 / 3.0).real
        # return np.nan_to_num(finternal, copy=False, nan=1.0).real

    return 1.0 + boost0 * _calc_finternal(r_norm)
    # return 1.0 + boost0 * (1 - _calc_finternal(r_norm)) / (r_norm**2 - 1)


def compute_powerlaw_boost(r_proj, r_scale, boost0=0.1, alpha=-1):
    r"""Computes the boost factor at radii r_proj using a power-law parametric form

    .. math::
        B(r_{\rm proj}) = 1 + B_0 \left(\frac{r_{\rm proj}}{r_{\rm scale}}\right)^\alpha

    Parameters
    ----------
    r_proj : array_like
        Radii
    r_scale : float
        Scale radius in same units as r_proj
    boost0 : float, optional
        Boost factor normalisation
    alpha : float, optional
        Exponent for the power-law parametrisation. NB: -1.0 (as in Melchior+16)

    Returns
    -------
    array
        Boost factor at each value of r_proj
    """

    r_norm = np.array(r_proj) / r_scale

    return 1.0 + boost0 * (r_norm) ** alpha


boost_models = {
    "nfw_boost": compute_nfw_boost,
    "powerlaw_boost": compute_powerlaw_boost,
}


def correct_with_boost_values(profile_vals, boost_factors):
    """Given a list of profile values (e.g., shear or DeltaSigma) and boost values,
    computes corrected profile

    Parameters
    ----------
    profile_vals : array_like
        Uncorrected profile values
    boost_factors : array_like
        Boost values, pre-computed

    Returns
    -------
    profile_vals_corrected : numpy.ndarray
        Corrected radial profile
    """

    profile_vals_corrected = np.array(profile_vals) / np.array(boost_factors)
    return profile_vals_corrected


def correct_with_boost_model(r_proj, profile_vals, boost_model, boost_rscale, **boost_model_kw):
    """Given a boost model and sigma profile, compute corrected sigma

    Parameters
    ----------
    r_proj : array_like
        Radii
    profile_vals : array_like
        Uncorrected profile values
    boost_model : str
        Boost model to use for correcting sigma
            * 'nfw_boost' - NFW profile model (Default)
            * 'powerlaw_boost' - Powerlaw profile
    boost_rscale : float
        Scale radius to be used with the boost model


    Returns
    -------
    profile_vals_corrected : numpy.ndarray
        Correted radial profile
    """
    boost_model_func = boost_models[boost_model]
    boost_factors = boost_model_func(r_proj, boost_rscale, **boost_model_kw)

    profile_vals_corrected = np.array(profile_vals) / boost_factors
    return profile_vals_corrected


boost_models = {
    "nfw_boost": compute_nfw_boost,
    "powerlaw_boost": compute_powerlaw_boost,
}



"""@file ccl.py
Cosmology using CCL
"""

import numpy as np
import pyccl as ccl

from ..utils import _patch_rho_crit_to_cd2018
from .parent_class import CLMMCosmology

__all__ = []


class CCLCosmology(CLMMCosmology):
    """
    Cosmology object

    Attributes
    ----------
    backend: str
        Name of back-end used
    be_cosmo: cosmology library
        Cosmology library used in the back-end
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # this tag will be used to check if the cosmology object is accepted by the modeling
        self.backend = "ccl"

        assert isinstance(self.be_cosmo, ccl.Cosmology)

        # cor factor for sigma_critical
        self.cor_factor = _patch_rho_crit_to_cd2018(ccl.physical_constants.RHO_CRITICAL)

    def _init_from_cosmo(self, be_cosmo):
        assert isinstance(be_cosmo, ccl.Cosmology)
        self.be_cosmo = be_cosmo

    def _init_from_params(self, H0, Omega_b0, Omega_dm0, Omega_k0):
        # pylint: disable=arguments-differ
        self.be_cosmo = ccl.Cosmology(
            Omega_c=Omega_dm0,
            Omega_b=Omega_b0,
            Omega_k=Omega_k0,
            h=H0 / 100.0,
            sigma8=0.8,
            n_s=0.96,
            T_CMB=2.7255,
            Neff=3.046,
            m_nu=[0.06, 0.0, 0.0],
            transfer_function="eisenstein_hu",
            matter_power_spectrum="linear",
        )

    def _set_param(self, key, value):
        raise NotImplementedError("CCL do not support changing parameters")

    def _get_param(self, key):
        if key == "Omega_m0":
            value = ccl.omega_x(self.be_cosmo, 1.0, "matter")
        elif key == "Omega_b0":
            value = self.be_cosmo["Omega_b"]
        elif key == "Omega_dm0":
            value = self.be_cosmo["Omega_c"]
        elif key == "Omega_k0":
            value = self.be_cosmo["Omega_k"]
        elif key == "h":
            value = self.be_cosmo["h"]
        elif key == "H0":
            value = self.be_cosmo["h"] * 100.0
        else:
            raise ValueError(f"Unsupported parameter {key}")
        return value

    def _get_Omega_m(self, z):
        a = self._get_a_from_z(z)
        return ccl.omega_x(self.be_cosmo, a, "matter")

    def _get_E2(self, z):
        a = self._get_a_from_z(z)
        return (ccl.h_over_h0(self.be_cosmo, a)) ** 2

    def _get_E2Omega_m(self, z):
        a = self._get_a_from_z(z)
        return ccl.omega_x(self.be_cosmo, a, "matter") * (ccl.h_over_h0(self.be_cosmo, a)) ** 2

    def _get_rho_m(self, z):
        # total matter density in physical units [Msun/Mpc3]
        a = self._get_a_from_z(z)
        return ccl.rho_x(self.be_cosmo, a, "matter", is_comoving=False)

    def _get_rho_c(self, z):
        a = self._get_a_from_z(z)
        return ccl.rho_x(self.be_cosmo, a, "critical", is_comoving=False)

    def _eval_da_z1z2_core(self, z1, z2):
        a1 = np.atleast_1d(self._get_a_from_z(z1))
        a2 = np.atleast_1d(self._get_a_from_z(z2))
        if len(a1) == 1 and len(a2) != 1:
            a1 = np.full_like(a2, a1)
        elif len(a2) == 1 and len(a1) != 1:
            a2 = np.full_like(a1, a2)

        da = ccl.angular_diameter_distance(self.be_cosmo, a1, a2)
        res = da if np.iterable(z1) or np.iterable(z2) else da.item()

        return res

    def _eval_sigma_crit_core(self, z_len, z_src):
        a_len = self._get_a_from_z(z_len)
        a_src = self._get_a_from_z(z_src)

        return self.be_cosmo.sigma_critical(a_lens=a_len, a_source=a_src) * self.cor_factor

    def _eval_linear_matter_powerspectrum(self, k_vals, redshift):
        return ccl.linear_matter_power(self.be_cosmo, k_vals, self._get_a_from_z(redshift))



"""General utility functions that are used in multiple modules"""

from astropy import units as u


def convert_units(dist1, unit1, unit2, redshift=None, cosmo=None):
    """Convenience wrapper to convert between a combination of angular and physical units.

    Supported units: radians, degrees, arcmin, arcsec, Mpc, kpc, pc
    (letter case independent)

    To convert between angular and physical units you must provide both
    `redshift` and a CLMM Cosmology object `cosmo`.

    Parameters
    ----------
    dist1 : float, array_like
        Input distances
    unit1 : str
        Unit for the input distances
    unit2 : str
        Unit for the output distances
    redshift : float, None, optional
        Redshift used to convert between angular and physical units. Default: None
    cosmo : clmm.Cosmology, None, optional
        CLMM Cosmology object to compute angular diameter distance to
        convert between physical and angular units. Default: None

    Returns
    -------
    dist2: float, numpy.ndarray
        Input distances converted to unit2
    """
    # make case independent
    unit1, unit2 = unit1.lower(), unit2.lower()
    # Available units
    angular_bank = {
        "radians": u.rad,
        "degrees": u.deg,
        "arcmin": u.arcmin,
        "arcsec": u.arcsec,
    }
    physical_bank = {"pc": u.pc, "kpc": u.kpc, "mpc": u.Mpc}
    units_bank = {**angular_bank, **physical_bank}
    # Some error checking
    if unit1 not in units_bank:
        raise ValueError(f"Input units ({unit1}) not supported")
    if unit2 not in units_bank:
        raise ValueError(f"Output units ({unit2}) not supported")
    # Try automated astropy unit conversion
    try:
        dist2 = (dist1 * units_bank[unit1]).to(units_bank[unit2]).value
    # Otherwise do manual conversion
    except u.UnitConversionError:
        # Make sure that we were passed a redshift and cosmology
        if redshift is None or cosmo is None:
            raise TypeError(
                "Redshift and cosmology must be specified to convert units"
            ) from u.UnitConversionError
        # Redshift must be greater than zero for this approx
        if not redshift > 0.0:
            raise ValueError("Redshift must be greater than 0.") from u.UnitConversionError
        # Convert angular to physical
        if (unit1 in angular_bank) and (unit2 in physical_bank):
            dist1_rad = (dist1 * units_bank[unit1]).to(u.rad).value
            dist1_mpc = cosmo.rad2mpc(dist1_rad, redshift)
            dist2 = (dist1_mpc * u.Mpc).to(units_bank[unit2]).value
        # Otherwise physical to angular
        else:
            dist1_mpc = (dist1 * units_bank[unit1]).to(u.Mpc).value
            dist1_rad = cosmo.mpc2rad(dist1_mpc, redshift)
            dist2 = (dist1_rad * u.rad).to(units_bank[unit2]).value
    return dist2
